from xasp.entities import Explain
from dumbo_asp.primitives.models import Model

explain = Explain.the_program(
    "a :- not b. b :- not a.",
    the_answer_set=Model.of_atoms("a"),
    the_atoms_to_explain=Model.of_atoms("a"),
)
explain.compute_explanation_dag()  # it can also be omitted
print(explain.explanation_dag())
explain.show_navigator_graph()


import base64
from enum import Enum, auto

import clingo
import clingo.ast
import typeguard
from clingo.ast import ASTType, ComparisonOperator, AggregateFunction, Sign, Location

from valid8 import validate


@typeguard.typechecked
class Transformer(clingo.ast.Transformer):
    def __init__(self) -> None:
        super().__init__()
        self.__called = False
        self.__result = []
        self.__exceptions = []
        self.__input = None

    def apply(self, string: str) -> str:
        validate("called once", self.__called, equals=False)
        self.__input = string.split('\n')
        clingo.ast.parse_string(string, lambda obj: self.__transform(obj))
        if self.__exceptions:
            raise self.__exceptions[0]
        return '\n'.join(self.__result)

    def __transform(self, obj):
        try:
            self.visit(obj)
        except Exception as err:
            self.__exceptions.append(err)

    def add_to_result(self, string: str) -> None:
        self.__result.append(string)

    def input(self, location: Location) -> str:
        res = []
        if location.begin.line == location.end.line:
            res.append(self.__input[location.begin.line - 1][location.begin.column - 1:location.end.column - 1])
        else:
            res.append(self.__input[location.begin.line - 1][location.begin.column - 1:])
            res.extend(self.__input[location.begin.line:location.end.line - 1])
            res.append(self.__input[location.end.line - 1][:location.end.column - 1])
        return '\n'.join(line.rstrip() for line in res if line.strip())

    def encode_input(self, location: Location) -> str:
        return base64.b64encode(self.input(location).encode()).decode()


@typeguard.typechecked
class ProgramSerializerTransformer(Transformer):
    class State(Enum):
        READING_HEAD = auto()
        READING_BODY = auto()
        READING_AGGREGATE = auto()

    def __init__(self) -> None:
        super().__init__()
        self.__rule_index = 0
        self.__agg_index = 0
        self.__state = None
        self.__variables = set()
        self.__definitions = []

    def visit_Definition(self, node):
        self.add_to_result(str(node))

    def visit_Rule(self, node):
        self.__rule_index += 1
        self.__state = self.State.READING_HEAD
        head = self.visit(node.head)
        self.__state = self.State.READING_BODY
        body = self.visit_sequence(node.body)
        variables = ','.join(sorted(self.__variables))
        rule_id = f"r{self.__rule_index}({variables})" if self.__variables else f"r{self.__rule_index}"
        rule_atom = f"rule({rule_id})"
        rule_body = self.__compute_rule_body(body)
        self.add_to_result(f"{rule_atom} :- {rule_body}.")

        self.add_to_result(f'original_rule(r{self.__rule_index},"{self.encode_input(node.location)}","{variables}").')

        if str(head) != "#false":
            validate("head type", head.ast_type, is_in=(ASTType.Aggregate, ASTType.Literal))
            if head.ast_type == ASTType.Aggregate:
                self.__process_choice(head, rule_id, rule_atom)
            else:
                validate("positive head", head.sign, equals=False)
                self.add_to_result(f"head({rule_id},{head}) :- {rule_atom}.")

        for literal in body:
            if literal.atom.ast_type == ASTType.Comparison:
                continue
            if literal.sign:
                validate("negation in front of atoms", literal.atom.ast_type, equals=ASTType.SymbolicAtom)
                if literal.atom.ast_type == ASTType.SymbolicAtom:
                    self.add_to_result(f"neg_body({rule_id},{literal.atom}) :- {rule_atom}.")
            elif literal.atom.ast_type == ASTType.SymbolicAtom:
                self.add_to_result(f"pos_body({rule_id},{literal.atom}) :- {rule_atom}.")
            elif literal.atom.ast_type == ASTType.BodyAggregate:
                self.__process_aggregate(literal, rule_id, rule_atom, variables)
            else:
                raise ValueError(f"Cannot process {literal}")

        self.__state = None
        self.__variables.clear()
        return node

    @staticmethod
    def __compute_rule_body(body, atom_predicate: str = "atom"):
        res = []
        for literal in body:
            if literal.atom.ast_type == ASTType.Comparison:
                res.append(f"{literal}")
            elif literal.sign == Sign.NoSign and literal.atom.ast_type == ASTType.SymbolicAtom:
                res.append(f"{atom_predicate}({literal.atom})")
        return ", ".join(res)

    @staticmethod
    def __compute_choice_bounds(choice):
        left, right = 0, "unbounded"
        if choice.left_guard is not None:
            validate("left guard", choice.left_guard.comparison != ComparisonOperator.NotEqual, equals=True)
            if choice.left_guard.comparison == ComparisonOperator.LessThan:
                left = f"{choice.left_guard.term} + 1"
            elif choice.left_guard.comparison == ComparisonOperator.LessEqual:
                left = f"{choice.left_guard.term}"
            elif choice.left_guard.comparison == ComparisonOperator.GreaterThan:
                right = f"{choice.left_guard.term} - 1"
            elif choice.left_guard.comparison == ComparisonOperator.GreaterEqual:
                right = f"{choice.left_guard.term}"
            elif choice.left_guard.comparison == ComparisonOperator.Equal:
                left = f"{choice.left_guard.term}"
                right = f"{choice.left_guard.term}"
            else:
                raise ValueError("Choice with != are not supported.")
        if choice.right_guard is not None:
            validate("right guard", choice.right_guard.comparison, is_in=[ComparisonOperator.LessThan,
                                                                          ComparisonOperator.LessEqual])
            if choice.right_guard.comparison == ComparisonOperator.LessThan:
                right = f"{choice.right_guard.term} + 1"
            elif choice.right_guard.comparison == ComparisonOperator.LessEqual:
                right = f"{choice.right_guard.term}"
        return left, right

    def __process_choice(self, head, rule_id, rule_atom):
        lower_bound, upper_bound = self.__compute_choice_bounds(head)
        self.add_to_result(f"choice({rule_id},{lower_bound},{upper_bound}) :- {rule_atom}.")
        if len(head.elements) > 1:
            for element in head.elements:
                validate("condition is empty", element.condition, length=0,
                         help_msg="Multiple elements are only supported without conditions")
                self.add_to_result(f"head({rule_id},{element.literal}) :- {rule_atom}.")
        elif len(head.elements) == 1:
            element = head.elements[0]
            if len(element.condition) == 0:
                self.add_to_result(f"head({rule_id},{element.literal}) :- {rule_atom}.")
            else:
                condition = self.__compute_rule_body(element.condition, "true")
                self.add_to_result(f"head({rule_id},{element.literal}) :- {rule_atom}, {condition}.")

    def __process_aggregate(self, literal, rule_id, rule_atom, variables):
        if literal.atom.right_guard is None:
            operator, bound = str(literal.atom.left_guard).split()
            if operator in ['>', '>=']:
                operator = operator.replace('>', '<')
            elif operator in ['<', '<=']:
                operator = operator.replace('<', '>')
        else:
            left = str(literal.atom.left_guard).split()
            right = str(literal.atom.right_guard).split()
            validate("left guard is < or <=", left[0][0], equals="<")
            validate("right guard is < or <=", right[0][0], equals="<")
            if left[0] == "<":
                left[1] = f"({left[1]}) + 1"
            if right[0] == "<":
                right[1] = f"({right[1]}) - 1"
            operator = "in"
            bound = f"({left[1]},{right[1]})"

        self.__agg_index += 1
        agg_id, fun = f"agg{self.__agg_index}", literal.atom.function
        if self.__variables:
            agg_id = f"{agg_id}({','.join(sorted(self.__variables))})"
        self.add_to_result(f"pos_body({rule_id},{agg_id}) :- {rule_atom}.")
        if fun == AggregateFunction.Sum:
            fun = "sum"
        elif fun == AggregateFunction.Count:
            fun = "count"
        elif fun == AggregateFunction.Min:
            fun = "min"
        elif fun == AggregateFunction.Max:
            fun = "max"
        else:
            raise ValueError(f"Cannot process aggregate function with ID {fun}")

        self.add_to_result(f'aggregate({agg_id},{fun},"{operator}",{bound}) :- {rule_atom}.')
        self.add_to_result(f'original_rule({agg_id},"{self.encode_input(literal.location)}","{variables}") :- {rule_atom}.')
        for element in literal.atom.elements:
            validate("condition", element.condition, length=1)
            validate("condition", element.condition[0].sign, equals=Sign.NoSign)
            validate("terms", element.terms, min_len=1)
            terms = [str(term) for term in element.terms]
            self.add_to_result(
                f"agg_set({agg_id},{element.condition[0]},{terms[0]},({','.join(terms[1:])})) :- "
                f"{rule_atom}, atom({element.condition[0]})."
            )

    def visit_Function(self, node):
        self.visit_children(node)
        return node

    def visit_BodyAggregate(self, node):
        self.__state = self.State.READING_AGGREGATE
        self.visit_children(node)
        self.__state = self.State.READING_BODY
        return node

    def visit_Variable(self, node):
        if self.__state == self.__state.READING_BODY:
            self.__variables.add(str(node))
        return node



from pathlib import Path

from xasp.commands import save_explanation_dag
from xasp.queries import compute_explanation_dag, compute_serialization





if __name__ == '__main__':
    true_atoms = [atom.strip() for atom in """
        edge(a,b)
        edge(a,d)
        edge(d,c)
        source(a)
        source(b)
        sink(c)
        reach(a,a)
        reach(b,b)
        reach(b,a)
        reach(a,d)
        reach(a,c)
        reach(b,d)
        reach(b,c)
        arc(b,a)
        arc(a,d)
        arc(d,c)
        threshold(0)
    """.split('\n')]
    serialization = compute_serialization("""
        1 <= {arc(X,Y); arc(Y,X)} <= 1 :- edge(X,Y).
        reach(X,X) :- source(X).
        reach(X,Y) :- reach(X,Z), arc(Z,Y).
        fail(X,Y) :- source(X), sink(Y), not reach(X,Y).
        :- threshold(T), #sum{1,X,Y : fail(X,Y)} > T.
        
        edge(a,b).
        edge(a,d).
        edge(d,c).
        
        source(a).
        source(b).
        
        sink(c).
        
        threshold(0).
    """, true_atoms=true_atoms, false_atoms=[atom.strip() for atom in """
        arc(a,b)
        arc(d,a)
        arc(c,d)
        fail(a,c)
        fail(b,c)
        reach(a,b)
        reach(c,a)
        reach(c,b)
        reach(c,c)
        reach(c,d)
        reach(d,a)
        reach(d,b)
        reach(d,c)
        reach(d,d)
    """.split('\n')], atom_to_explain="arc(a,b)")
    dag = compute_explanation_dag(serialization)
    save_explanation_dag(dag, answer_set=true_atoms, atom="arc(a,b)", target=Path("a.png"))



from typing import Optional, Any

import clingo
import clingo.ast
import typeguard

from dumbo_asp.primitives.models import Model
from dumbo_asp.primitives.predicates import Predicate
from xasp.entities import Explain
from dumbo_utils.primitives import PositiveIntegerOrUnbounded


@typeguard.typechecked
def compute_stable_model(asp_program: str, context: Optional[Any] = None) -> Optional[Model]:
    control = clingo.Control()
    control.add("base", [], asp_program)
    control.ground([("base", [])], context=context)
    try:
        return Model.of_control(control)
    except Model.NoModelError:
        return None


@typeguard.typechecked
def compute_serialization(asp_program: str, answer_set: Model, additional_atoms_in_base: Model = Model.empty(),
                          atoms_to_explain: Model = Model.empty()) -> Model:
    return Explain.the_program(
        asp_program,
        the_answer_set=answer_set,
        the_atoms_to_explain=atoms_to_explain,
        the_additional_atoms_in_the_base=additional_atoms_in_base,
    ).serialization.drop(Predicate.parse("original_rule"))


@typeguard.typechecked
def process_aggregates(to_be_explained_serialization: Model) -> Model:
    explain = Explain.the_serialization(
        to_be_explained_serialization
    )
    explain.process_aggregates()
    return explain.serialization.drop(Predicate.parse("original_rule"))


@typeguard.typechecked
def compute_atoms_explained_by_initial_well_founded(serialization: Model) -> Model:
    return Explain.the_serialization(
        serialization
    ).atoms_explained_by_initial_well_founded


@typeguard.typechecked
def compute_minimal_assumption_set(to_be_explained_serialization: Model) -> Model:
    return Explain.the_serialization(to_be_explained_serialization).minimal_assumption_set()


@typeguard.typechecked
def compute_minimal_assumption_sets(
        to_be_explained_serialization: Model,
        atoms_to_explain: Model,
        up_to: Optional[int] = None
) -> tuple[Model, ...]:
    explain = Explain.the_serialization(
        to_be_explained_serialization,
        the_atoms_to_explain=atoms_to_explain,
    )
    explain.compute_minimal_assumption_set(
        repeat=up_to if up_to is not None else PositiveIntegerOrUnbounded.of_unbounded()
    )
    return tuple(explain.minimal_assumption_set(index) for index in range(explain.minimal_assumption_sets))


@typeguard.typechecked
def compute_explanation(to_be_explained_serialization: Model) -> Model:
    return Explain.the_serialization(to_be_explained_serialization).explanation_sequence()


@typeguard.typechecked
def compute_explanations(
        to_be_explained_serialization: Model,
        atoms_to_explain: Model,
        up_to: Optional[int] = None
) -> tuple[Model, ...]:
    explain = Explain.the_serialization(
        to_be_explained_serialization,
        the_atoms_to_explain=atoms_to_explain,
    )
    explain.compute_explanation_sequence(
        repeat=up_to if up_to is not None else PositiveIntegerOrUnbounded.of_unbounded()
    )
    return tuple(explain.explanation_sequence(index) for index in range(explain.explanation_sequences))


@typeguard.typechecked
def compute_explanation_dag(to_be_explained_serialization: Model) -> Model:
    return Explain.the_serialization(to_be_explained_serialization).explanation_dag()


@typeguard.typechecked
def compute_explanation_dags(
        to_be_explained_serialization: Model,
        atoms_to_explain: Model,
        up_to: Optional[int] = None
) -> tuple[Model, ...]:
    explain = Explain.the_serialization(
        to_be_explained_serialization,
        the_atoms_to_explain=atoms_to_explain,
    )
    explain.compute_explanation_dag(
        repeat=up_to if up_to is not None else PositiveIntegerOrUnbounded.of_unbounded()
    )
    return tuple(explain.explanation_dag(index) for index in range(explain.explanation_dags))



from pathlib import Path
from typing import Callable, Final, Any


PROJECT_ROOT: Final = Path(__file__).parent.parent


def call_with_difference_if_invalid_index(index: int, length: int, callback: Callable[[int], Any]):
    if index >= 0:
        if length <= index:
            callback(index - length + 1)
    else:
        if length < -index:
            callback(-index - length)



from pathlib import Path
from typing import Optional
from dumbo_asp.primitives.models import Model

from xasp.entities import Explain


def save_explanation_dag(dag: Model, answer_set: Model, atoms_to_explain: Model, target: Path,
                         distance: Optional[int] = None,
                         **kwargs):
    Explain.the_dag(
        dag,
        the_answer_set=answer_set,
        the_atoms_to_explain=atoms_to_explain,
    ).save_igraph(target, distance, **kwargs)




import dataclasses
from collections import defaultdict, namedtuple
from functools import cached_property

import typeguard
from clingo import Number

from dumbo_utils.console import log


@typeguard.typechecked
class ProcessAggregatesContext:
    @staticmethod
    def check_operator(operator, bounds, value):
        if operator.string in ["=", "=="]:
            return Number(1) if value == bounds else Number(0)
        if operator.string in ["!=", "<>"]:
            return Number(1) if value != bounds else Number(0)
        if operator.string == "<":
            return Number(1) if value < bounds else Number(0)
        if operator.string == ">":
            return Number(1) if value > bounds else Number(0)
        if operator.string == "<=":
            return Number(1) if value <= bounds else Number(0)
        if operator.string == ">=":
            return Number(1) if value >= bounds else Number(0)
        if operator.string == "in":
            return Number(1) if bounds.arguments[0] <= value <= bounds.arguments[1] else Number(0)
        assert False


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class ComputeWellFoundedContext:
    Rule = namedtuple("Rule", "id head pos_body neg_body")
    WellFoundedModel = namedtuple("WellFoundedModel", "true potentially_true")

    rules: dict = dataclasses.field(default_factory=dict)
    atom2heads: dict = dataclasses.field(default_factory=lambda: defaultdict(list))
    atom2pos_bodies: dict = dataclasses.field(default_factory=lambda: defaultdict(list))
    atom2neg_bodies: dict = dataclasses.field(default_factory=lambda: defaultdict(list))

    def collect_rule(self, rule):
        assert rule not in self.rules.keys()
        self.rules[rule] = self.Rule(rule, set(), set(), set())
        return Number(1)

    def collect_head(self, rule, atom):
        assert rule in self.rules.keys()
        rule = self.rules[rule]
        rule.head.add(atom)
        self.atom2heads[atom].append(rule)
        return Number(1)

    def collect_pos_body(self, rule, atom):
        assert rule in self.rules.keys()
        rule = self.rules[rule]
        rule.pos_body.add(atom)
        self.atom2pos_bodies[atom].append(rule)
        return Number(1)

    def collect_neg_body(self, rule, atom):
        assert rule in self.rules.keys()
        rule = self.rules[rule]
        rule.neg_body.add(atom)
        self.atom2neg_bodies[atom].append(rule)
        return Number(1)

    def false_in_well_founded_model(self, atom):
        return Number(1) if atom not in self.well_founded_model.potentially_true else Number(0)

    @cached_property
    def well_founded_model(self) -> 'ComputeWellFoundedContext.WellFoundedModel':
        log.debug("Compute well-founded model: begin")
        well_founded_model = self.WellFoundedModel(
            set(),
            set().union(*(rule.head for rule in self.rules.values() if rule.head))
        )
        source_pointer = {atom: None for atom in well_founded_model.potentially_true}

        while True:
            queue = list(set().union(*(rule.head for rule in self.rules.values() if
                                       len(rule.pos_body) + len(rule.neg_body) == 0 and rule.head)))
            while queue:
                true_atom = queue[-1]
                queue.pop()

                if true_atom in well_founded_model.true:
                    continue
                assert true_atom in well_founded_model.potentially_true

                well_founded_model.true.add(true_atom)
                log.debug(f"founded: {true_atom}")
                for rule in self.atom2pos_bodies[true_atom]:
                    if all(atom in well_founded_model.true for atom in rule.pos_body) and \
                            all(atom not in well_founded_model.potentially_true for atom in rule.neg_body):
                        for head_atom in rule.head:
                            queue.append(head_atom)

            queue = []
            for rule in self.rules.values():
                self.enqueue_if_source_pointer(rule, queue, source_pointer, well_founded_model)
            while queue:
                atom_with_source_pointer = queue[-1]
                queue.pop()

                for rule in self.atom2pos_bodies[atom_with_source_pointer]:
                    self.enqueue_if_source_pointer(rule, queue, source_pointer, well_founded_model)
            founded_atoms = set(atom for atom in source_pointer if source_pointer[atom] is not None)
            unfounded_atoms = well_founded_model.potentially_true - founded_atoms
            if not unfounded_atoms:
                break

            for unfounded_atom in unfounded_atoms:
                log.debug(f"unfounded: {unfounded_atom}")
                assert unfounded_atom not in well_founded_model.true
                well_founded_model.potentially_true.remove(unfounded_atom)
        log.debug("Compute well-founded model: end")
        return well_founded_model

    @staticmethod
    def enqueue_if_source_pointer(rule, my_queue, source_pointer, well_founded_model):
        if rule.head is None:
            return
        if all(source_pointer[head_atom] is not None for head_atom in rule.head):
            return
        log.debug(f"check {rule.id}")
        if any(atom not in well_founded_model.potentially_true for atom in rule.pos_body):
            return
        log.debug(f"  pos_body potentially true")
        if any(atom in well_founded_model.true for atom in rule.neg_body):
            return
        log.debug(f"  neg_body potentially true")
        if all(source_pointer[atom] is not None for atom in rule.pos_body):
            for head_atom in rule.head:
                if source_pointer[head_atom] is None:
                    log.debug(f"source pointer: {head_atom} {rule.id}")
                    source_pointer[head_atom] = rule
                    my_queue.append(head_atom)


@dataclasses.dataclass(frozen=True)
class ComputeExplanationContext:
    __index: list[int] = dataclasses.field(default_factory=lambda: [0], init=False)

    def index(self):
        self.__index[0] += 1
        return Number(self.__index[0])



import base64
import dataclasses
import json
import webbrowser
import zlib
from dataclasses import InitVar
from enum import auto, IntEnum
from pathlib import Path
from typing import Callable, Final, Optional, Dict, List, Any, Union

import clingo
import igraph
import typeguard
from clingo import Model
from dumbo_asp.primitives.models import Model
from dumbo_asp.primitives.predicates import Predicate
from dumbo_asp.primitives.atoms import GroundAtom
from dumbo_utils.url import compress_object_for_url
from valid8 import validate

from xasp.contexts import ComputeExplanationContext, ProcessAggregatesContext, ComputeWellFoundedContext
from dumbo_utils.primitives import PositiveIntegerOrUnbounded
from xasp.transformers import ProgramSerializerTransformer
from xasp.utils import call_with_difference_if_invalid_index


@typeguard.typechecked
@dataclasses.dataclass
class Explain:
    key: InitVar[Any]
    __key = object()

    __state: "Explain.State" = dataclasses.field(default_factory=lambda: Explain.State.INITIAL, init=False)
    __commands_implementation: Dict[str, Callable] = dataclasses.field(default_factory=dict)
    __asp_program: Optional[str] = dataclasses.field(default=None, init=False)
    __answer_set: Optional[Model] = dataclasses.field(default=None, init=False)
    __additional_atoms_in_the_base: Optional[Model] = dataclasses.field(default=None, init=False)
    __atoms_to_explain: Optional[Model] = dataclasses.field(default=None, init=False)
    __serialization: Model = dataclasses.field(default=Model.empty(), init=False)
    __atoms_explained_by_initial_well_founded: Model = dataclasses.field(default=Model.empty(), init=False)
    __minimal_assumption_sets: List[Model] = dataclasses.field(default_factory=list, init=False)
    __minimal_assumption_sets_block_constraints: List[str] = dataclasses.field(default_factory=list, init=False)
    __explanation_sequences: List[Model] = dataclasses.field(default_factory=list, init=False)
    __explanation_dags: List[Model] = dataclasses.field(default_factory=list, init=False)
    __igraph: List[Optional[igraph.Graph]] = dataclasses.field(default_factory=list, init=False)

    class State(IntEnum):
        INITIAL = auto()
        SERIALIZED = auto()
        AGGREGATE_PROCESSED = auto()
        WELL_FOUNDED_COMPUTED = auto()
        MINIMAL_ASSUMPTION_SET_COMPUTED = auto()
        EXPLANATION_SEQUENCE_COMPUTED = auto()
        EXPLANATION_DAG_COMPUTED = auto()
        IGRAPH_COMPUTED = auto()

    def __post_init__(self, key):
        validate("key", key, equals=self.__key, help_msg="Use a factory method")

    @staticmethod
    def the_program(
            value: str,
            the_answer_set: Model,
            the_atoms_to_explain: Model = Model.empty(),
            the_additional_atoms_in_the_base: Model = Model.empty()
    ) -> "Explain":
        res = Explain(key=Explain.__key)
        res.__asp_program = value
        res.__answer_set = the_answer_set
        res.__atoms_to_explain = the_atoms_to_explain
        res.__additional_atoms_in_the_base = the_additional_atoms_in_the_base
        res.__compute_serialization()
        return res

    @staticmethod
    def the_serialization(
            value: Model,
            the_answer_set: Optional[Model] = None,
            the_atoms_to_explain: Optional[Model] = None,
            the_additional_atoms_in_the_base: Model = Model.empty(),
    ) -> "Explain":
        res = Explain(key=Explain.__key)
        res.__serialization = value
        res.__answer_set = the_answer_set
        res.__atoms_to_explain = the_atoms_to_explain
        res.__additional_atoms_in_the_base = the_additional_atoms_in_the_base
        res.__state = Explain.State.SERIALIZED
        return res

    @staticmethod
    def the_dag(
            value: Model,
            the_answer_set: Optional[Model] = None,
            the_atoms_to_explain: Optional[Model] = None,
            the_additional_atoms_in_the_base: Model = Model.empty(),
    ) -> "Explain":
        res = Explain(key=Explain.__key)
        res.__explanation_dags.append(value)
        res.__state = Explain.State.SERIALIZED
        res.__answer_set = the_answer_set
        res.__atoms_to_explain = the_atoms_to_explain
        res.__additional_atoms_in_the_base = the_additional_atoms_in_the_base
        res.__state = Explain.State.EXPLANATION_DAG_COMPUTED
        return res

    def process_aggregates(self) -> None:
        if self.__state >= Explain.State.AGGREGATE_PROCESSED:
            return
        if self.__state < Explain.State.SERIALIZED:
            self.__compute_serialization()
        validate("state", self.__state, equals=Explain.State.SERIALIZED)
        self.__serialization = self.__process_aggregates()
        self.__state = Explain.State.AGGREGATE_PROCESSED

    def compute_atoms_explained_by_initial_well_founded(self) -> None:
        if self.__state >= Explain.State.WELL_FOUNDED_COMPUTED:
            return
        if self.__state < Explain.State.AGGREGATE_PROCESSED:
            self.process_aggregates()
        validate("state", self.__state, equals=Explain.State.AGGREGATE_PROCESSED)
        self.__atoms_explained_by_initial_well_founded = self.__compute_atoms_explained_by_initial_well_founded()
        self.__state = Explain.State.WELL_FOUNDED_COMPUTED

    def compute_minimal_assumption_set(self, repeat: Union[int, PositiveIntegerOrUnbounded] = 1) -> None:
        if type(repeat) is int:
            repeat = PositiveIntegerOrUnbounded.of(repeat)
        if self.__state < Explain.State.WELL_FOUNDED_COMPUTED:
            self.compute_atoms_explained_by_initial_well_founded()
        validate("state", self.__state, min_value=Explain.State.WELL_FOUNDED_COMPUTED)
        repeat += len(self.__minimal_assumption_sets)
        while repeat.greater_than(len(self.__minimal_assumption_sets)):
            assumption_set = self.__compute_minimal_assumption_set()
            if assumption_set is None:
                break
            self.__minimal_assumption_sets.append(assumption_set)
        self.__state = max(self.__state, Explain.State.MINIMAL_ASSUMPTION_SET_COMPUTED)

    def compute_explanation_sequence(self, repeat: Union[int, PositiveIntegerOrUnbounded] = 1) -> None:
        if type(repeat) is int:
            repeat = PositiveIntegerOrUnbounded.of(repeat)
        if self.__state < Explain.State.MINIMAL_ASSUMPTION_SET_COMPUTED:
            self.compute_minimal_assumption_set()
        validate("state", self.__state, min_value=Explain.State.MINIMAL_ASSUMPTION_SET_COMPUTED)
        repeat += len(self.__explanation_sequences)
        while repeat.greater_than(len(self.__explanation_sequences)):
            explanation = self.__compute_explanation_sequence()
            if explanation is not None:
                self.__explanation_sequences.append(explanation)
            else:
                assumption_sets = len(self.__minimal_assumption_sets)
                self.compute_minimal_assumption_set()
                if len(self.__minimal_assumption_sets) == assumption_sets:
                    break
        self.__state = max(self.__state, Explain.State.EXPLANATION_SEQUENCE_COMPUTED)

    def compute_explanation_dag(self, repeat: Union[int, PositiveIntegerOrUnbounded] = 1) -> None:
        if type(repeat) is int:
            repeat = PositiveIntegerOrUnbounded.of(repeat)
        if self.__state < Explain.State.EXPLANATION_SEQUENCE_COMPUTED:
            self.compute_explanation_sequence()
        validate("state", self.__state, min_value=Explain.State.EXPLANATION_SEQUENCE_COMPUTED)
        repeat += len(self.__explanation_dags)
        while repeat.greater_than(len(self.__explanation_dags)):
            dag = self.__compute_explanation_dag()
            if dag is not None:
                self.__explanation_dags.append(dag)
            else:
                sequences = len(self.__explanation_sequences)
                self.compute_explanation_sequence()
                if len(self.__explanation_sequences) == sequences:
                    break
        self.__state = Explain.State.EXPLANATION_DAG_COMPUTED

    def compute_igraph(self, index: int = -1) -> None:
        validate("answer_set", self.__answer_set, help_msg="Answer set was not provided")
        validate("atoms_to_explain", self.__atoms_to_explain, help_msg="Atoms to explain were not provided")
        validate("additional_atoms_in_the_base", self.__additional_atoms_in_the_base,
                 help_msg="Additional atoms were not provided")
        validate("must have a DAG", self.explanation_dag(index), help_msg="No DAG with the given index")
        while len(self.__igraph) < self.explanation_dags:
            self.__igraph.append(None)
        if self.__igraph[index] is None:
            self.__igraph[index] = self.__compute_igraph(dag=self.__explanation_dags[index])
        self.__state = max(self.__state, Explain.State.IGRAPH_COMPUTED)

    def save_igraph(self, filename: Path, index: int = -1, **kwargs) -> None:
        self.compute_igraph(index)
        igraph.plot(
            self.__igraph[index],
            layout=self.__igraph[index].layout_sugiyama(),
            margin=140,
            target=filename,
            vertex_label_dist=2,
            vertex_size=8,
            **kwargs,
        )

    def show_navigator_graph(self, index: int = -1) -> None:
        self.compute_igraph(index)
        url = "https://xasp-navigator.alviano.net/#"
        # url = "http://localhost:5173/#"
        url += compress_object_for_url(self.navigator_graph(index))
        webbrowser.open(url, new=0, autoraise=True)

    @property
    def asp_program(self) -> Optional[str]:
        return self.__asp_program

    @property
    def answer_set(self) -> Optional[Model]:
        return self.__answer_set

    @property
    def atoms_to_explain(self) -> Optional[Model]:
        return self.__atoms_to_explain

    @property
    def additional_atoms_in_the_base(self) -> Optional[Model]:
        return self.__additional_atoms_in_the_base

    @property
    def serialization(self) -> Model:
        validate("state", self.__state, min_value=Explain.State.SERIALIZED)
        return self.__serialization

    @property
    def atoms_explained_by_initial_well_founded(self) -> Model:
        if self.__state < Explain.State.WELL_FOUNDED_COMPUTED:
            self.compute_atoms_explained_by_initial_well_founded()
        validate("state", self.__state, min_value=Explain.State.WELL_FOUNDED_COMPUTED)
        return self.__atoms_explained_by_initial_well_founded

    @property
    def minimal_assumption_sets(self) -> int:
        return len(self.__minimal_assumption_sets)

    def minimal_assumption_set(self, index: int = -1) -> Model:
        call_with_difference_if_invalid_index(index, self.minimal_assumption_sets, self.compute_minimal_assumption_set)
        return self.__minimal_assumption_sets[index]

    @property
    def explanation_sequences(self) -> int:
        return len(self.__explanation_sequences)

    def explanation_sequence(self, index: int = -1) -> Model:
        call_with_difference_if_invalid_index(index, self.explanation_sequences, self.compute_explanation_sequence)
        return self.__explanation_sequences[index]

    @property
    def explanation_dags(self) -> int:
        return len(self.__explanation_dags)

    def explanation_dag(self, index: int = -1) -> Model:
        call_with_difference_if_invalid_index(index, self.explanation_dags, self.compute_explanation_dag)
        return self.__explanation_dags[index]

    def navigator_graph(self, index: int = -1) -> Dict:
        self.compute_igraph(index)
        graph = self.__igraph[index]
        layout = graph.layout_sugiyama()
        res = {
            "nodes": [
                {
                    "id": index,
                    "label": node.attributes()["label"],
                    "x": layout.coords[index][0],
                    "y": layout.coords[index][1],
                }
                for index, node in enumerate(graph.vs)
            ],
            "links": [
                {
                    "source": link.tuple[0],
                    "target": link.tuple[1],
                    "label": link.attributes()["label"],
                }
                for link in graph.es
            ],
        }
        return res

    @staticmethod
    def compute_stable_model(asp_program: str, context: Optional[Any] = None) -> Optional[Model]:
        control = clingo.Control()
        control.add("base", [], asp_program)
        control.ground([("base", [])], context=context)
        try:
            return Model.of_control(control)
        except Model.NoModelError:
            return None
    
    def __compute_serialization(self) -> None:
        validate("state", self.__state, equals=Explain.State.INITIAL)

        strongly_negated_atoms = {str(atom)[1:] for atom in self.answer_set if str(atom).startswith('-')}
        strongly_negated_atoms.update(str(atom)[1:] for atom in self.additional_atoms_in_the_base
                                      if str(atom).startswith('-'))
        strongly_negated_atoms.update(str(atom)[1:] for atom in self.atoms_to_explain if str(atom).startswith('-'))

        transformer = ProgramSerializerTransformer()
        transformed_program = transformer.apply(self.asp_program + '\n'.join(f":- {atom}, -{atom}."
                                                                             for atom in strongly_negated_atoms))
        model = self.compute_stable_model(
            SERIALIZATION_ENCODING + transformed_program +
            '\n'.join(f"true({atom})." for atom in self.answer_set) +
            '\n'.join(f"atom({atom})." for atom in self.additional_atoms_in_the_base) +
            '\n'.join(f"explain({atom})." for atom in self.atoms_to_explain)
        )
        self.__serialization = model
        self.__state = max(self.__state, Explain.State.SERIALIZED)

    def __process_aggregates(self) -> Model:
        res = self.compute_stable_model(
            PROCESS_AGGREGATES_ENCODING + self.serialization.as_facts,
            context=ProcessAggregatesContext()
        )
        validate("res", res, help_msg="No stable model. The input is likely wrong.")
        return res

    def __compute_atoms_explained_by_initial_well_founded(self) -> Model:
        encoding = WELL_FOUNDED_ENCODING + self.serialization.as_facts
        return self.compute_stable_model(encoding, context=ComputeWellFoundedContext())

    def __compute_minimal_assumption_set(self) -> Optional[Model]:
        while len(self.__minimal_assumption_sets_block_constraints) < len(self.__minimal_assumption_sets):
            validate("can enumerate", self.atoms_to_explain, max_len=1,
                     help_msg="At most one atom to explain must be passed to the factory method")
            mas = self.__minimal_assumption_sets[len(self.__minimal_assumption_sets_block_constraints)]
            self.__minimal_assumption_sets_block_constraints.append(mas.block_up)
            if len(self.atoms_to_explain) > 0 and len(self.__minimal_assumption_sets_block_constraints) == 1:
                atom = f"assume_false({self.atoms_to_explain[0]})"
                self.__minimal_assumption_sets_block_constraints[0] += \
                    f"\n:- {'not ' if atom in (str(x) for x in mas) else ''}{atom}."
        encoding = MINIMAL_ASSUMPTION_SET_ENCODING + EXPLAIN_ENCODING + \
                   self.serialization.as_facts + \
                   self.atoms_explained_by_initial_well_founded.as_facts + \
                   '\n'.join(constraint for constraint in self.__minimal_assumption_sets_block_constraints)
        res = self.compute_stable_model(encoding)
        if not self.__minimal_assumption_sets:
            validate("res", res, help_msg="No stable model. The input is likely wrong.")
        return res

    def __compute_explanation_sequence(self) -> Optional[Model]:
        instance: Final = self.minimal_assumption_set(-1).as_facts + \
                          self.serialization.as_facts + \
                          self.atoms_explained_by_initial_well_founded.as_facts
        encoding = EXPLANATION_ENCODING + EXPLAIN_ENCODING + instance + \
                   '\n'.join(model.project(Predicate.parse("explained_by/3"), 1).block_up
                             for model in self.__explanation_sequences)
        res = self.compute_stable_model(encoding, context=ComputeExplanationContext())

        if res is None:
            validate("must have an explanation", self.__explanation_sequences, min_len=1,
                     help_msg="No stable model. The input is likely wrong.")
            return None

        encoding = INDEXED_EXPLAIN_ENCODING + instance + res.as_facts
        res = self.compute_stable_model(encoding, context=ComputeExplanationContext())
        assert res is not None

        def fun(atom: GroundAtom) -> GroundAtom:
            fun.index += 1
            return GroundAtom(clingo.Function(atom.predicate_name,
                                              [clingo.Number(fun.index)] + list(atom.arguments[1:])))

        fun.index = 0

        return res.map(fun)

    def __compute_explanation_dag(self) -> Optional[Model]:
        encoding = EXPLANATION_DAG_ENCODING + self.serialization.as_facts + \
                   self.__explanation_sequences[-1].as_facts + \
                   '\n'.join(model.filter(lambda atom: atom.arguments[-1].type != clingo.SymbolType.String)
                             .substitute(Predicate.parse("link/2"), 1, clingo.Function("_")).block_up
                             for model in self.__explanation_dags)
        res = self.compute_stable_model(encoding, context=ComputeExplanationContext())
        if not self.__explanation_dags:
            validate("res", res, help_msg="No stable model. The input is likely wrong.")
        return res

    def __compute_igraph(self, dag: Model) -> igraph.Graph:
        graph = igraph.Graph(directed=True)

        rules = {}
        for rule in dag.drop(Predicate.parse("link/4")):
            validate("predicate", rule.predicate_name, equals="original_rule")
            rule_index = str(rule.arguments[0])
            b64 = rule.arguments[1].string
            variables = rule.arguments[2].string
            rules[rule_index] = (base64.b64decode(b64).decode(), variables)

        for link in dag.drop(Predicate.parse("original_rule/3")):
            validate("link name", link.predicate_name, equals="link")
            source = str(link.arguments[1])
            label = link.arguments[2]
            sink = str(link.arguments[3])
            reason = self.__link_reason(rules, label).split('\n', maxsplit=1)
            color = GRAPH_COLOR[reason[0]]
            if len(graph.vs) == 0 or len(graph.vs.select(name=source)) == 0:
                graph.add_vertex(source, color=color, label=f"{source}\n{reason[0]}")
            if sink not in ['"true"', '"false"']:
                if sink == '"#true"' and len(graph.vs.select(name=sink)) == 0:
                    graph.add_vertex(sink, color=color, label=f"#true")
                validate(f"{sink} is present", graph.vs.select(name=sink), length=1)
                graph.add_edge(source, sink, color=color, label=reason[1] if len(reason) > 1 else None)

        if len(self.__atoms_to_explain) == 0:
            return graph

        reachable_nodes = graph.neighborhood(
            vertices=[str(atom) for atom in self.__atoms_to_explain],
            order=len(graph.vs),
            mode="out"
        )
        nodes = []
        for reachable_nodes_element in reachable_nodes:
            nodes.extend(reachable_nodes_element)
        return graph.induced_subgraph(list(set(nodes)))

    @staticmethod
    def __link_reason(rules, label: clingo.Symbol) -> str:
        if label.name in ["assumption", "initial_well_founded"]:
            return label.name.replace('_', ' ')
        validate("name", label.name, equals="")
        validate("arguments", label.arguments, length=2)
        rule, variables = rules[
            label.arguments[1].name if label.arguments[1].name.startswith("r") else
            str(label.arguments[1]) if label.arguments[1].name.startswith("agg") else
            str(label.arguments[1].arguments[0])

        ]
        return f"{label.arguments[0].name.replace('_', ' ')}\n{rule}" + \
            (f"\n{variables} => {','.join(str(x) for x in label.arguments[1].arguments)}"
             if label.arguments[1].arguments else "")


GRAPH_COLOR: Final = {
    "support": "#90EE90",  # lightgreen
    "assumption": "#800080",  # purple
    "initial well founded": "#8B0000",  # darkred
    "lack of support": "#FF0000",  # red
    "choice rule": "#F08080",  # lightcoral
    "required to falsify body": "#FF8C00",  # dark orange
}

PROCESS_AGGREGATES_ENCODING: Final = """
%******************************************************************************
Enrich the representation of a program with aggregates so that minimal assumption sets can be computed with respect to 
a program without aggregates.


__RUN__

$ cat input.asp | clingo /dev/stdin process_aggregates.asp --outf=1 | sed -n '6 p'


__INPUT FORMAT__

Each rule of the program is encoded by facts of the form
- rule(RULE_ID)
- original_rule(RULE_INDEX, BASE64, VARIABLES)
- choice(RULE_ID, LOWER_BOUND, UPPER_BOUND)
- head(RULE_ID, ATOM)
- pos_body(RULE_ID, ATOM|AGGREGATE)
- neg_body(RULE_ID, ATOM)

Each aggregate of the program is encoded by facts of the form
- aggregate(AGG, FUN, OPERATOR, BOUNDS)
- agg_set(AGG, ATOM, WEIGHT, TERMS)

The answer set is encoded by facts of the form
- true(ATOM)
- false(ATOM)

Atoms to explain, which cannot be assumed false are encoded by
- explain(ATOM)

******************************************************************************%

% compute true aggregates
true_aggregate(Agg) :- aggregate(Agg, Fun, Operator, Bounds), Fun == sum;
    Value = #sum{Weight, Terms : agg_set(Agg, Atom, Weight, Terms), true(Atom)};
    @check_operator(Operator, Bounds, Value) = 1.
true_aggregate(Agg) :- aggregate(Agg, Fun, Operator, Bounds), Fun == count;
    Value = #count{Weight, Terms : agg_set(Agg, Atom, Weight, Terms), true(Atom)};
    @check_operator(Operator, Bounds, Value) = 1.
true_aggregate(Agg) :- aggregate(Agg, Fun, Operator, Bounds), Fun == min;
    Value = #min{Weight, Terms : agg_set(Agg, Atom, Weight, Terms), true(Atom)};
    @check_operator(Operator, Bounds, Value) = 1.
true_aggregate(Agg) :- aggregate(Agg, Fun, Operator, Bounds), Fun == max;
    Value = #max{Weight, Terms : agg_set(Agg, Atom, Weight, Terms), true(Atom)};
    @check_operator(Operator, Bounds, Value) = 1.

% every aggregate that is not true, is false
false_aggregate(Agg) :- aggregate(Agg, Fun, Operator, Bounds); not true_aggregate(Agg).

% true aggregates are considered as rules of the form  agg :- true_atoms_in_agg_set, ~false_atoms_in_agg_set.
rule(Agg) :- true_aggregate(Agg).
head(Agg,Agg) :- true_aggregate(Agg).
pos_body(Agg,Atom) :- true_aggregate(Agg), agg_set(Agg,Atom,Weight,Terms), true(Atom).
neg_body(Agg,Atom) :- true_aggregate(Agg), agg_set(Agg,Atom,Weight,Terms), false(Atom).

% false aggregates are considered as several rules of the form  agg :- ~true_atom_in_agg_set.   agg :- false_atom_in_agg_set.
rule((Agg,Atom)) :- false_aggregate(Agg), agg_set(Agg,Atom,Weight,Terms).
head((Agg,Atom),Agg) :- false_aggregate(Agg), agg_set(Agg,Atom,Weight,Terms).
pos_body((Agg,Atom),Atom) :- false_aggregate(Agg), agg_set(Agg,Atom,Weight,Terms), false(Atom).
neg_body((Agg,Atom),Atom) :- false_aggregate(Agg), agg_set(Agg,Atom,Weight,Terms), true(Atom).


#show.
#show rule/1.
#show original_rule/3.
#show choice/3.
#show head/2.
#show pos_body/2.
#show neg_body/2.
#show true/1.
#show false/1.
#show explain/1.
#show aggregate(Agg) : aggregate(Agg, Fun, Operator, Bounds).
#show true(Agg) : true_aggregate(Agg).
#show false(Agg) : false_aggregate(Agg).


% avoid warnings
rule(0) :- #false.
original_rule(0,0,0) :- #false.
choice(0,0,0) :- #false.
head(0,0) :- #false.
pos_body(0,0) :- #false.
neg_body(0,0) :- #false.
aggregate(0,0,0,0) :- #false.
agg_set(0,0,0,0) :- #false.
true(0) :- #false.
false(0) :- #false.
explain(0) :- #false.
"""

EXPLAIN_ENCODING: Final = """
%******************************************************************************
__INPUT FORMAT__

Each rule of the program is encoded by facts of the form
- rule(RULE_ID)
- head(RULE_ID, ATOM)
- pos_body(RULE_ID, ATOM|AGGREGATE)
- neg_body(RULE_ID, ATOM)
- choice(RULE_ID, LOWER_BOUND, UPPER_BOUND)

Aggregates are identified by facts of the form
- aggregate(AGGREGATE)

The answer set is encoded by facts of the form
- true(ATOM|AGGREGATE)
- false(ATOM|AGGREGATE)

Atoms explained by initial well founded are encoded by facts of the form
- explained_by(ATOM, initial_well_founded)

******************************************************************************%

% all atoms need to be explained by exactly one reason
atom(Atom) :- true(Atom).
atom(Atom) :- false(Atom).
:- atom(Atom), #count{Reason: explained_by(Atom,Reason)} != 1.
has_explanation(Atom) :- explained_by(Atom,_).


% assumed false atoms are explained (by assumption)
explained_by(Atom, assumption) :- assume_false(Atom).


% true atoms can be explained by a supporting rule whose body literals already have an explanation
{explained_by(Atom, (support, Rule))} :- 
  true(Atom);
  head(Rule,Atom);
  true(BAtom) : pos_body(Rule,BAtom);
  has_explanation(BAtom) : pos_body(Rule,BAtom);
  false(BAtom) : neg_body(Rule,BAtom);
  has_explanation(BAtom) : neg_body(Rule,BAtom).


% explain false atoms : begin

    % false atoms can be explained if all the possibly supporting rules already have an explanation
    {explained_by(Atom, lack_of_support)} :-
      false(Atom);
      false_body(Rule) : head(Rule,Atom).

    % a non-supporting rule is explained if there is some false body literal that already has an explanation
    false_body(Rule) :-
      rule(Rule);
      pos_body(Rule,BAtom), false(BAtom), has_explanation(BAtom).
    false_body(Rule) :-
      rule(Rule);
      neg_body(Rule,BAtom), true(BAtom), has_explanation(BAtom).


    % a false atom can be explained by a rule with false head and whose body contains the false atom, and all other body literals are true
    {explained_by(Atom, (required_to_falsify_body, Rule))} :-
      false(Atom), not aggregate(Atom);
      pos_body(Rule,Atom), false_head(Rule);
      true(BAtom) : pos_body(Rule,BAtom), BAtom != Atom;
      has_explanation(BAtom) : pos_body(Rule,BAtom), BAtom != Atom;
      false(BAtom) : neg_body(Rule,BAtom);
      has_explanation(BAtom) : neg_body(Rule,BAtom).
    explained_head(Rule) :-
      rule(Rule);
      has_explanation(HAtom) : head(Rule,HAtom).
    false_head(Rule) :- 
      explained_head(Rule), not choice(Rule,_,_);
      false(HAtom) : head(Rule,HAtom).
    false_head(Rule) :-
      explained_head(Rule), choice(Rule, LowerBound, UpperBound); 
      not LowerBound <= #count{HAtom : head(Rule,HAtom), true(HAtom)} <= UpperBound.

    % a false atom can be explained by a choice rule with true body and whose true head atoms already reach the upper bound
    {explained_by(Atom, (choice_rule, Rule))} :-
      false(Atom);
      head(Rule,Atom), choice(Rule, LowerBound, UpperBound), UpperBound != unbounded;
      true(BAtom) : pos_body(Rule,BAtom);
      has_explanation(BAtom) : pos_body(Rule,BAtom);
      false(BAtom) : neg_body(Rule,BAtom);
      has_explanation(BAtom) : neg_body(Rule,BAtom);
      #count{HAtom : head(Rule, HAtom), true(HAtom), has_explanation(HAtom)} = UpperBound.

% explain false atoms : end


% avoid warnings
rule(0) :- #false.
choice(0,0,0) :- #false.
head(0,0) :- #false.
pos_body(0,0) :- #false.
neg_body(0,0) :- #false.
aggregate(0) :- #false.
true(0) :- #false.
false(0) :- #false.
explained_by(0,initial_well_founded) :- #false.
"""

INDEXED_EXPLAIN_ENCODING: Final = """
%******************************************************************************
__INPUT FORMAT__

Each rule of the program is encoded by facts of the form
- rule(RULE_ID)
- head(RULE_ID, ATOM)
- pos_body(RULE_ID, ATOM|AGGREGATE)
- neg_body(RULE_ID, ATOM)
- choice(RULE_ID, LOWER_BOUND, UPPER_BOUND)

Aggregates are identified by facts of the form
- aggregate(AGGREGATE)

Atoms to explain are encoded by
- explain(ATOM)

The answer set is encoded by facts of the form
- true(ATOM|AGGREGATE)
- false(ATOM|AGGREGATE)

******************************************************************************%

has_explanation(Atom) :- indexed_explained_by(_,Atom,_).

indexed_explained_by(@index(), Atom, assumption) :- assume_false(Atom).
indexed_explained_by(@index(), Atom, initial_well_founded) :- false(Atom), explained_by(Atom, initial_well_founded).

% true atoms can be explained by a supporting rule whose body literals already have an explanation
indexed_explained_by(@index(), Atom, (support, Rule)) :-
  explained_by(Atom, (support, Rule));
  true(Atom);
  head(Rule,Atom);
  true(BAtom) : pos_body(Rule,BAtom);
  has_explanation(BAtom) : pos_body(Rule,BAtom);
  false(BAtom) : neg_body(Rule,BAtom);
  has_explanation(BAtom) : neg_body(Rule,BAtom).


% explain false atoms : begin

    % false atoms can be explained if all the possibly supporting rules already have an explanation
    indexed_explained_by(@index(), Atom, lack_of_support) :-
      explained_by(Atom, lack_of_support);
      false(Atom);
      false_body(Rule) : head(Rule,Atom).

    % a non-supporting rule is explained if there is some false body literal that already has an explanation
    false_body(Rule) :-
      rule(Rule);
      pos_body(Rule,BAtom), false(BAtom), has_explanation(BAtom).
    false_body(Rule) :-
      rule(Rule);
      neg_body(Rule,BAtom), true(BAtom), has_explanation(BAtom).


    % a false atom can be explained by a rule with false head and whose body contains the false atom, and all other body literals are true
    indexed_explained_by(@index(), Atom, (required_to_falsify_body, Rule)) :-
      explained_by(Atom, (required_to_falsify_body, Rule));
      false(Atom), not aggregate(Atom);
      pos_body(Rule,Atom), false_head(Rule);
      true(BAtom) : pos_body(Rule,BAtom), BAtom != Atom;
      has_explanation(BAtom) : pos_body(Rule,BAtom), BAtom != Atom;
      false(BAtom) : neg_body(Rule,BAtom);
      has_explanation(BAtom) : neg_body(Rule,BAtom).
    explained_head(Rule) :-
      rule(Rule);
      has_explanation(HAtom) : head(Rule,HAtom).
    false_head(Rule) :-
      explained_head(Rule), not choice(Rule,_,_);
      false(HAtom) : head(Rule,HAtom).
    false_head(Rule) :-
      explained_head(Rule), choice(Rule, LowerBound, UpperBound);
      not LowerBound <= #count{HAtom : head(Rule,HAtom), true(HAtom)} <= UpperBound.

    % a false atom can be explained by a choice rule with true body and whose true head atoms already reach the upper bound
    indexed_explained_by(@index(), Atom, (choice_rule, Rule)) :-
      explained_by(Atom, (choice_rule, Rule));
      false(Atom);
      head(Rule,Atom), choice(Rule, LowerBound, UpperBound), UpperBound != unbounded;
      true(BAtom) : pos_body(Rule,BAtom);
      has_explanation(BAtom) : pos_body(Rule,BAtom);
      false(BAtom) : neg_body(Rule,BAtom);
      has_explanation(BAtom) : neg_body(Rule,BAtom);
      #count{HAtom : head(Rule, HAtom), true(HAtom), has_explanation(HAtom)} = UpperBound.

% explain false atoms : end


% keep only atoms connected to the query : begin

    relevant(Atom) :- explain(Atom).
    relevant(Atom) :- not explain(_); indexed_explained_by(_, Atom, _).

    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (support, Rule));
        pos_body(Rule, Atom').
    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (support, Rule));
        neg_body(Rule, Atom').

    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (lack_of_support));
        head(Rule, Atom);
        pos_body(Rule, Atom'), false(Atom'), indexed_explained_by(Index', Atom', _), Index' < Index.
    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (lack_of_support));
        head(Rule, Atom);
        neg_body(Rule, Atom'), true(Atom'), indexed_explained_by(Index', Atom', _), Index' < Index.

    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (required_to_falsify_body, Rule));
        head(Rule, Atom').
    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (required_to_falsify_body, Rule));
        pos_body(Rule, Atom').
    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (required_to_falsify_body, Rule));
        neg_body(Rule, Atom').

    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (choice_rule, Rule));
        head(Rule, Atom'), true(Atom').
    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (choice_rule, Rule));
        pos_body(Rule, Atom').
    relevant(Atom') :-
        relevant(Atom), indexed_explained_by(Index, Atom, (choice_rule, Rule));
        neg_body(Rule, Atom').

% keep only atoms connected to the query : end

#show.
#show explained_by(Index, Atom, Reason) : indexed_explained_by(Index, Atom, Reason), relevant(Atom).

% avoid warnings
rule(0) :- #false.
choice(0,0,0) :- #false.
head(0,0) :- #false.
pos_body(0,0) :- #false.
neg_body(0,0) :- #false.
aggregate(0) :- #false.
explain(0) :- #false.
true(0) :- #false.
false(0) :- #false.
explained_by(0,0) :- #false.
assume_false(0) :- #false.
"""

MINIMAL_ASSUMPTION_SET_ENCODING: Final = """
%******************************************************************************
Compute minimal assumption sets for a program wrt. an answer set.
If the atom to explain is false, the considered assumption sets will not assume its falsity.

__INPUT FORMAT__

Everything from EXPLAIN_ENCODING.

Atoms to explain, which cannot be assumed false are encoded by
- explain(ATOM)

******************************************************************************%


% guess the assumption set and minimize its cardinality
%   aggregates cannot be assumed false
%   atoms to explain should not be assumed false
{assume_false(Atom)} :- false(Atom), not aggregate(Atom).
:~ false(Atom), assume_false(Atom), not explain(Atom). [1@1, Atom]
:~ false(Atom), assume_false(Atom), explain(Atom). [1@2, Atom]


#show.
#show assume_false/1.

% avoid warnings
explain(0) :- #false.
"""

EXPLANATION_ENCODING: Final = """
%******************************************************************************
Compute explanation for a program wrt. an answer set and a minimal assumption set.

__INPUT FORMAT__

Everything from EXPLAIN_ENCODING.

For each atom in the minimal assumption set, the input must contain an atom of the form
- assume_false(ATOM)

******************************************************************************%


#show.
#show explained_by/2.

% avoid warnings
assume_false(0) :- #false.
"""

EXPLANATION_DAG_ENCODING: Final = """
%******************************************************************************
Compute explanation for a program wrt. an answer set and a minimal assumption set.

__INPUT FORMAT__

Each rule of the program is encoded by facts of the form
- rule(RULE_ID)
- original_rule(RULE_INDEX, BASE64, VARIABLES)
- head(RULE_ID, ATOM)
- pos_body(RULE_ID, ATOM|AGGREGATE)
- neg_body(RULE_ID, ATOM)
- choice(RULE_ID, LOWER_BOUND, UPPER_BOUND)

Aggregates are identified by facts of the form
- aggregate(AGGREGATE)

The answer set is encoded by facts of the form
- true(ATOM|AGGREGATE)
- false(ATOM|AGGREGATE)

For each atom in the answer set, the input must contain an atom of the form
- explained_by(INDEX, ATOM, REASON)
  where REASON is one of
  - assumption
  - initial_well_founded
  - (support, RULE)
  - lack_of_support
  - (required_to_falsify_body, Rule)
  - (choice_rule, Rule)
******************************************************************************%

link(Index, Atom, Reason, BAtom) :- explained_by(Index, Atom, Reason);
    Reason = (support, Rule);
    pos_body(Rule, BAtom).
link(Index, Atom, Reason, BAtom) :- explained_by(Index, Atom, Reason);
    Reason = (support, Rule);
    neg_body(Rule, BAtom).

{
    link(Index, Atom, (lack_of_support, Rule), BAtom) : 
        pos_body(Rule, BAtom), false(BAtom), explained_by(Index', BAtom, _), Index' < Index;
    link(Index, Atom, (lack_of_support, Rule), BAtom) : 
        neg_body(Rule, BAtom), true (BAtom), explained_by(Index', BAtom, _), Index' < Index
} = 1 :- 
    explained_by(Index, Atom, Reason);
    Reason = lack_of_support;
    head(Rule, Atom).

link(Index, Atom, Reason, HAtom) :- explained_by(Index, Atom, Reason);
    Reason = (required_to_falsify_body, Rule);
    head(Rule, HAtom).
link(Index, Atom, Reason, BAtom) :- explained_by(Index, Atom, Reason);
    Reason = (required_to_falsify_body, Rule);
    pos_body(Rule, BAtom), BAtom != Atom.
link(Index, Atom, Reason, BAtom) :- explained_by(Index, Atom, Reason);
    Reason = (required_to_falsify_body, Rule);
    neg_body(Rule, BAtom).

link(Index, Atom, Reason, HAtom) :- explained_by(Index, Atom, Reason);
    Reason = (choice_rule, Rule);
    head(Rule, HAtom), true(HAtom).
link(Index, Atom, Reason, BAtom) :- explained_by(Index, Atom, Reason);
    Reason = (choice_rule, Rule);
    pos_body(Rule, BAtom).
link(Index, Atom, Reason, BAtom) :- explained_by(Index, Atom, Reason);
    Reason = (choice_rule, Rule);
    neg_body(Rule, BAtom).

#show.
#show link(Index, Atom, Reason, Atom') : link(Index, Atom, Reason, Atom').
#show link(Index, Atom, Reason, "true") : explained_by(Index, Atom, Reason), true(Atom),
    #count{Atom' : link(Index, Atom, _, Atom')} = 0.
#show link(Index, Atom, Reason, "#true") : explained_by(Index, Atom, Reason), true(Atom),
    #count{Atom' : link(Index, Atom, _, Atom')} = 0, Reason = (support, Rule), choice(Rule, _, _).
#show link(Index, Atom, Reason, "false") : explained_by(Index, Atom, Reason), false(Atom),
    #count{Atom' : link(Index, Atom, _, Atom')} = 0.
#show original_rule/3.

% avoid warnings
rule(0) :- #false.
original_rule(0,0,0) :- #false.
choice(0,0,0) :- #false.
head(0,0) :- #false.
pos_body(0,0) :- #false.
neg_body(0,0) :- #false.
aggregate(0) :- #false.
true(0) :- #false.
false(0) :- #false.
explained_by(0,0,0) :- #false.
"""

SERIALIZATION_ENCODING: Final = """
atom(Atom) :- explain(Atom).
atom(Atom) :- true(Atom).
atom(Atom) :- head(Rule, Atom).
atom(Atom) :- pos_body(Rule, Atom), not aggregate(Atom,_,_,_).
atom(Atom) :- neg_body(Rule, Atom).
atom(Atom) :- agg_set(Aggregate, Atom, Weight, Terms).

false(Atom) :- atom(Atom), not true(Atom).

#show.
#show rule/1.
#show original_rule/3.
#show choice/3.
#show head/2.
#show pos_body/2.
#show neg_body/2.
#show aggregate/4.
#show agg_set/4.
#show true/1.
#show false/1.
#show explain/1.

% avoid warnings
rule(0) :- #false.
original_rule(0,0,0) :- #false.
head(0,0) :- #false.
choice(0,0,0) :- #false.
pos_body(0,0) :- #false.
neg_body(0,0) :- #false.
aggregate(0,0,0,0) :- #false.
agg_set(0,0,0,0) :- #false.
true(0) :- #false.
atom(0) :- #false.
explain(0) :- #false.
"""

WELL_FOUNDED_ENCODING: Final = """
%******************************************************************************
__INPUT FORMAT__

Each rule of the program is encoded by facts of the form
- rule(RULE_ID)
- head(RULE_ID, ATOM)
- pos_body(RULE_ID, ATOM|AGGREGATE)
- neg_body(RULE_ID, ATOM)

******************************************************************************%

collecting_rules :- rule(Rule), @collect_rule(Rule) != 1.
collecting_heads :- not collecting_rules, head(Rule,Atom), @collect_head(Rule,Atom) != 1.
collecting_pos_bodies :- not collecting_rules, pos_body(Rule,Atom), @collect_pos_body(Rule,Atom) != 1.
collecting_neg_bodies :- not collecting_rules, neg_body(Rule,Atom), @collect_neg_body(Rule,Atom) != 1.
collected_program :- not collecting_rules, not collecting_heads, not collecting_pos_bodies, not collecting_neg_bodies.

explained_by(Atom, initial_well_founded) :- collected_program; false(Atom), @false_in_well_founded_model(Atom) == 1.

#show.
#show explained_by/2.

% avoid warnings
rule(0) :- #false.
head(0,0) :- #false.
pos_body(0,0) :- #false.
neg_body(0,0) :- #false.
false(0) :- #false.
"""
